/*****************************************************************************
 * RRDtool 1.0.14  Copyright Tobias Oetiker, 1997 - 2000
 *****************************************************************************
 * rrd_stats.c  Startup wrapper
 *****************************************************************************/

#include "rrd_tool.h"
#include "rrd_statsl.h"

void PrintUsage(void);
int CountArgs(char *aLine);
int CreateArgs(char *, char *, int, char **);
int HandleInputLine(int, char **, FILE*);
#define TRUE		1
#define FALSE		0
#define MAX_LENGTH	10000


void PrintUsage(void)
{
    printf("\n"
		"RRDtool 1.0.24  Copyright 1997-2000 by Tobias Oetiker <tobi@oetiker.ch>\n\n"
		" statistics command-line utility\n\n"
		" usage: (4 function are implemented, perl-type call is described)\n"
		" ----------------------------------------------------------------\n"
		" mean/max/min: computes the average / minimum / maxmimum / integration value for a rrd database, for each ds\n"
		" my ($step, $start, $end, $res, $names) = \n"
		" 	RRDs::stats $rrd, \"func:pstep:cf\",\"--start=stime\",\"--end=etime\";\n"
		" parameters:\n"
		"  func : mean/max/min/int\n"
		"  pstep : proposed step (function finds the best based on time-range and rra/step availability)\n"
		"  cf : consolidation function\n"
		"  stime-etime : measure time\n"
		" results: \n"
		"  $step : the step that was found and choosen\n"
		"  $start, $end : the really start and end\n"
		"  $res : list of result\n"
		"  $names : list of corresponding ds names\n"
		" examples: \n"
		"  my ($step, $start, $end, $avg, $names) = RRDs::stats $rrd, \"mean:300:AVERAGE\",\"--start=-1m\";\n"
		"  my ($step, $start, $end, $avg, $names) = RRDs::stats $rrd, \"min:300:AVERAGE\",\"--start=-1d\";\n"
		"  my ($step, $start, $end, $avg, $names) = RRDs::stats $rrd, \"max:300:AVERAGE\",\"--start=-1w\";\n"
		"  my ($step, $start, $end, $avg, $names) = RRDs::stats $rrd, \"int:300:AVERAGE\",\"--start=-1w\";\n"
		" ----------------------------------------------------------------\n"
		" dist: computes the distribution of values for a given ds in an rrd\n"
		" my ($step, $start, $end, $before, $after, $res) = \n"
		" 	RRDs::stats $rrd, \"dist:minval:maxval:ds:nbdist:pstep:cf\",\"--start=stime\",\"--end=$etime\";\n"
		" parameters:\n"
		"  pstep : proposed step (function funds the best based on time-range and rra/step availability)\n"
		"  ds : name of ds\n"
		"  nbdist : number of interval to compute dist (between minval and maxval)\n"
		"  cf : consolidation function\n"
		"  stime-etime : measure time\n"
		"  maxval, minval : minimum and maximum value to compute the distribution\n"
		" results: \n"
		"  $step : the step that was found and choosen\n"
		"  $start, $end : the really start and end\n"
		"  $res : list of results\n"
		"  $before, $after : results that are <minval(before) of > maxval(after)\n"
		" example: \n"
		" my ($step, $sta, $en, $before, $after, $res) = RRDs::stats $rrd, \"dist:0:100:ds1:10:300:AVERAGE\",\"--start=-1d\";\n"
		" ----------------------------------------------------------------\n"
		" rrd_stats choose the rra with the following rules : (not same as fetch or graph)\n"
		" - for each rra it computes an overlap factor that is the factor between the requested \n"
		"   range (end-start) and the available range of this rra\n"
		" - for overlap factor that are not 'very' different (90%), it take the step that is \n"
		"   closest from the proposed step\n" );

}


int main(int argc, char *argv[])
{
    char **myargv;
    char aLine[MAX_LENGTH];
#ifdef MUST_DISABLE_SIGFPE
    signal(SIGFPE,SIG_IGN);
#endif
#ifdef MUST_DISABLE_FPMASK
    fpsetmask(0);
#endif
    if (argc == 1)
	{
	    PrintUsage();
	    return 0;
	}

    if ((argc == 2) && (*argv[1] == '-'))
	{
#if HAVE_GETRUSAGE
	  struct rusage  myusage;
	  struct timeval starttime;
	  struct timeval currenttime;
	  struct timezone tz;

	    tz.tz_minuteswest =0;
	    tz.tz_dsttime=0;
	    gettimeofday(&starttime,&tz);
#endif

	    while (fgets(aLine, sizeof(aLine)-1, stdin)){
		if ((argc = CountArgs(aLine)) == 0)  {
		    fprintf(stderr,"ERROR: not enough arguments\n");
		}
		if ((myargv = (char **) malloc((argc+1) *
					       sizeof(char *))) == NULL)   {
		    perror("malloc");
		    return -1;
		}
		if ((argc=CreateArgs(argv[0], aLine, argc, myargv)) < 0) {
		    fprintf(stderr, "ERROR: creating arguments\n");
		    return -1;
		}

		if (HandleInputLine(argc, myargv, stdout))
		    return -1;
		free(myargv);

#if HAVE_GETRUSAGE
		getrusage(RUSAGE_SELF,&myusage);
		gettimeofday(&currenttime,&tz);
		printf("OK u:%1.2f s:%1.2f r:%1.2f\n",
		       (double)myusage.ru_utime.tv_sec+
		       (double)myusage.ru_utime.tv_usec/1000000.0,
		       (double)myusage.ru_stime.tv_sec+
		       (double)myusage.ru_stime.tv_usec/1000000.0,
		       (double)(currenttime.tv_sec-starttime.tv_sec)
		       +(double)(currenttime.tv_usec-starttime.tv_usec)
		       /1000000.0);
#else
		printf("OK\n");
#endif
		fflush(stdout); /* this is important for pipes to work */
	    }
	}
    else
	HandleInputLine(argc, argv, stderr);
    return 0;
}

int HandleInputLine(int argc, char **argv, FILE* out)
{
    optind=0; /* reset gnu getopt */
    opterr=0; /* no error messages */

    if (argc < 3
	|| strcmp("help", argv[1]) == 0
	|| strcmp("--help", argv[1]) == 0
	|| strcmp("-help", argv[1]) == 0
	|| strcmp("-?", argv[1]) == 0
	|| strcmp("-h", argv[1]) == 0 ) {
	PrintUsage();
	return 0;
    }

    if (strcmp("stats", argv[1]) == 0) {
		struct stats_values *psv, *remind_psv;;

		if (rrd_stats (argc-1, &argv[1], &psv) == 1 ) {
			remind_psv = psv;
			while ( psv != NULL ) {
				if ( psv->nbvals > 0 ) {
					int ii;
					printf ( "[%i] multiple values [%c] ", psv->nbvals, psv->datatype );
					switch (psv->datatype) {
					case 'I':	{
									int * results = (int *) psv->val.ptr;
									for (ii=0; ii < psv->nbvals; ii++ ) {
										printf ( psv->format, results[ii] );
										if (ii < psv->nbvals-1) printf ( ", ");
									}
									break;
								}
					case 'S':	{
									char ** str = (char **) psv->val.ptr;
									for (ii=0; ii < psv->nbvals; ii++ ) {
										printf ( psv->format, str[ii] );
										if (ii < psv->nbvals-1) printf ( ", ");
									}
									break;
								}
					case 'D':	{
									double * results = (double *) psv->val.ptr;
									for (ii=0; ii < psv->nbvals; ii++ ) {
										printf ( psv->format, results[ii] );
										if (ii < psv->nbvals-1) printf ( ", ");
									}
									break;
								}
					default: break;
					}
					printf ( "\n" );
				} else {
					printf ( "single value [%c] ", psv->datatype );
					switch (psv->datatype) {
					case 'i':	printf ( psv->format, psv->val.i ); break;
					case 's':	printf ( psv->format, psv->val.str ); break;
					case 'd':	printf ( psv->format, psv->val.d ); break;
					default: break;
					}
					printf ( "\n" );
				}
				psv = psv->pnext;
			}
			free_psv ( remind_psv );
		}
	} else {
		rrd_set_error("unknown function '%s'",argv[1]);
    }
    if (rrd_test_error()) {
	fprintf(out, "ERROR: %s\n",rrd_get_error());
	rrd_clear_error();
    }
    return(0);
}

int CountArgs(char *aLine)
{
    int i=0;
    int aCount = 0;
    int inarg = 0;
    while (aLine[i] == ' ') i++;
    while (aLine[i] != 0){
	if((aLine[i]== ' ') && inarg){
	    inarg = 0;
	}
	if((aLine[i]!= ' ') && ! inarg){
	    inarg = 1;
	    aCount++;
	}
	i++;
    }
    return aCount;
}

/*
 * CreateArgs - take a string (aLine) and tokenize
 */
int CreateArgs(char *pName, char *aLine, int argc, char **argv)
{
    char	*getP, *putP;
    char	**pargv = argv;
    char        Quote = 0;
    int inArg = 0;
    int	len;

    len = strlen(aLine);
    /* remove trailing space and newlines */
    while (len && aLine[len] <= ' ') {
	aLine[len] = 0 ; len--;
    }
    /* sikp leading blanks */
    while (*aLine && *aLine <= ' ') aLine++;

    pargv[0] = pName;
    argc = 1;
    getP = aLine;
    putP = aLine;
    while (*getP){
	switch (*getP) {
	case ' ':
	    if (Quote){
		*(putP++)=*getP;
	    } else
		if(inArg) {
		    *(putP++) = 0;
		    inArg = 0;
		}
	    break;
	case '"':
	case '\'':
	    if (Quote != 0) {
		if (Quote == *getP)
		    Quote = 0;
		else {
		    *(putP++)=*getP;
		}
	    } else {
		if(!inArg){
		    pargv[argc++] = putP;
		    inArg=1;
		}
		Quote = *getP;
	    }
	    break;
	default:
	    if(!inArg){
		pargv[argc++] = putP;
		inArg=1;
	    }
	    *(putP++)=*getP;
	    break;
	}
	getP++;
    }

    *putP = '\0';

    if (Quote)
	return -1;
    else
	return argc;
}


